/*
 * srluaz.c
 * Lua interpreter for self-running programs containg lua or zip files.
 * See comment above main() for some usage info.
 *
 * Frank Bechmann <fBechmann(at)web(dot)de>
 * Last changed: 01/04/05 18:53:18
 * Based on srlua.c (http://luaforge.net/projects/srlua/) by:
 * Luiz Henrique de Figueiredo <lhf@tecgraf.puc-rio.br>
 * 08 Nov 2004 15:30:04
 * This code is hereby placed in the public domain.
 */

/* in most libs its WIN32, but here its _WIN32, so just map */
#if defined WIN32 && !defined(_WIN32)
#   define _WIN32
#endif

#ifdef _WIN32
#include <windows.h>
#endif

#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "glue.h"
#include "lua.h"
#include "lualib.h"
#include "lauxlib.h"

#include <unzip.h>
#ifdef _WIN32
#define USEWIN32IOAPI
#include <iowin32.h>
#endif

/*===  default file names searched in zip file during start up ===*/

#ifndef LUA_INIT_DEFAULT
# define LUA_INIT_DEFAULT "@init.lua"
#endif
#ifndef LUA_EXEC_DEFAULT
# define LUA_EXEC_DEFAULT "exec.lua"
#endif
#ifndef LUA_INIT_CFGFILE
# define LUA_INIT_CFGFILE "config.lua"
#endif

/*===  log and error handling macros =============================*/

/* log file name and line, than integer status and message string */
#define LLOG(L, status, msg) \
    fprintf(stderr, "\n[%s/%d]: %d (%s)", __FILE__, __LINE__, status, msg)

/* log file name and line, than lua stack */
#define PRINTSTACK(L) \
    printstack(L, __FILE__, __LINE__)

/* log file name and line, than message strings, finally exit */
#define FATAL(pgmn, L, msg1, msg2)  \
    message(pgmn, L, msg1, msg2, 1, __FILE__, __LINE__)

/* log file name and line, than message strings, continue after this */
#define MESSAGE(pgmn, L, msg1, msg2)  \
    message(pgmn, L, msg1, msg2, 0, __FILE__, __LINE__)

/* if call of cmd results in integer != 0, log via MESSAGE() and set status to result */
#define CHECK0(pgmn, L, status, cmd) \
    (errno=0,_check0(pgmn, L, &status, (cmd), #cmd, __FILE__, __LINE__))

/* if call of cmd results in pointer == NULL, log via MESSAGE() and set status to -1 or errno */
#define CHECKNN(pgmn, L, status, cmd) \
    (errno=0,_checkNN(pgmn, L, &status, (cmd), #cmd, __FILE__, __LINE__))

/* if evaluation of lua chunk results in integer != 0, log via message() */
#define DOSTRING(pgmn, L, s, chunk) \
    dostring(pgmn, L, s, chunk, __FILE__, __LINE__)

/* if evaluation of lua file results in integer != 0, log via message() */
#define DOFILE(pgmn, L, fn) \
    dofile(pgmn, L, fn, __FILE__, __LINE__)

/*===  structs ===================================================*/

typedef struct
{
    FILE *f;
    long size;
    char buff[512];
} State;

/*===  functions =================================================*/

static void openlibs(lua_State *L)
{
    int t=lua_gettop(L);
    luaopen_base(L);
    luaopen_table(L);
    luaopen_io(L);
    luaopen_string(L);
    luaopen_math(L);
    luaopen_debug(L);
    luaopen_loadlib(L);
    /* add your libraries here */
    lua_settop(L,t);
}

static void printstack(lua_State *L, const char* srcn, int line)
{
    int n=lua_gettop(L);
    int i;
    fprintf(stderr, "\n=== [%s/%d]", srcn, line);
    for (i=1; i<=n; i++) {
        if (lua_isstring(L,i)) fprintf(stderr, "\n[%d]: %s",i,lua_tostring(L,i));
        else if (lua_isnil(L,i)) fprintf(stderr, "\n[%d]: %s",i,"nil");
        else if (lua_isboolean(L,i)) fprintf(stderr, "\n[%d]: %s",i,lua_toboolean(L,i) ? "true" : "false");
        else fprintf(stderr, "\n[%d]: %s:%p",i,lua_typename(L,lua_type(L,i)),lua_topointer(L,i));
    }
    fprintf(stderr, "\n=== %d entries\n", n);
}

static void message(const char* pgmn, lua_State *L,
    const char* msg1, const char* msg2, int fatal, const char* srcn, int line)
{
    int save_errno = errno;
    int msglen = 32 + strlen(srcn)
                + (msg1 ? strlen(msg1) : 0) + (msg2 ? strlen(msg2) : 0)
                + (save_errno ? strlen(strerror(save_errno)) : 0);
    char* msgbuf = malloc(msglen);
    const char* msgptr = msg1 ? msg1 : (msg2 ? msg2 : srcn);
    if (msgbuf) {
        int msgoffset = sprintf(msgbuf, "[%s/%d]", srcn, line);
        if (msg1) msgoffset += sprintf(msgbuf+msgoffset, " %s", msg1);
        if (msg2) msgoffset += sprintf(msgbuf+msgoffset, " (%s)", msg2);
        if (save_errno) msgoffset += sprintf(msgbuf+msgoffset, " [%d,%s]", save_errno, strerror(save_errno));
        msgptr = msgbuf;
    }
    fprintf(stderr,"\n%s: %s%s",pgmn,msgptr,fatal?" (EXIT)":"");
    if (L) printstack(L, srcn, line);
#ifdef _WIN32
    MessageBox(NULL,msgptr,pgmn, MB_ICONERROR | MB_OK);
#endif
    if (fatal) {
        if (L)
            luaL_error(L,"\n%s: %s",pgmn,msgptr); /* this will exit() */
        else
            exit(save_errno ? save_errno : 1);
    }
    if (msgbuf) free(msgbuf);
    errno = save_errno;
}

static int _check0(const char* pgmn, lua_State *L,
    int* curr_status, int new_status, const char* msg, const char* srcn, int line)
{
    if (new_status) {
        char buf[16];
        message(pgmn, L, msg, itoa(new_status, buf, 10), 0, srcn, line);
        if (0 == *curr_status) *curr_status = new_status;
    }
    return new_status;
}

static void* _checkNN(const char* pgmn, lua_State *L,
    int* curr_status, void* ptr, const char* msg, const char* srcn, int line)
{
    if (NULL == ptr) {
        message(pgmn, L, msg, NULL, 0, srcn, line);
        if (0 == *curr_status) *curr_status = errno ? errno : -1;
    }
    return ptr;
}

static int report(const char* pgmn, lua_State *L,
    int status, const char* msg2, const char* srcn, int line)
{
  if (status) {
    const char* msg1 = lua_tostring(L, -1);
    if (msg1 == NULL) msg1 = "(error with no message)";
    message(pgmn, L, msg1, msg2, 0, srcn, line);
    lua_pop(L, 1);
  }
  return status;
}

static int lcall(const char* pgmn, lua_State *L, int narg, int clear)
{
  int status;
  int base = lua_gettop(L) - narg;  /* function index */
  lua_pushliteral(L, "_TRACEBACK");
  lua_rawget(L, LUA_GLOBALSINDEX);  /* get traceback function */
  lua_insert(L, base);              /* put it under chunk and args */
  status = lua_pcall(L, narg, (clear ? 0 : LUA_MULTRET), base);
  lua_remove(L, base);              /* remove traceback function */
  return status;
}

static int docall(const char* pgmn, lua_State *L,
    int status, const char* msg2, const char* srcn, int line)
{ return report(pgmn, L, status || lcall(pgmn, L, 0, 1), msg2, srcn, line); }

static int dostring(const char* pgmn, lua_State *L,
    const char* s, const char* chunk, const char* srcn, int line)
{ return docall(pgmn, L, luaL_loadbuffer(L, s, strlen(s), chunk), s, srcn, line); }

static int dofile(const char* pgmn, lua_State *L,
    char *fn, const char* srcn, int line)
{ return docall(pgmn, L, luaL_loadfile(L, fn), fn, srcn, line); }

static const char* myget(lua_State *L, void *data, size_t *size)
{
    State* s=data;
    size_t n;
    if (s->size<=0) return NULL;
    n=(sizeof(s->buff)<=s->size) ? sizeof(s->buff) : s->size;
    n=fread(s->buff,1,n,s->f);
    if (n==-1) return NULL;
    s->size-=n;
    *size=n;
    return s->buff;
}

/* return length of given file */
static size_t myfilelen(FILE* f)
{
    size_t len = 0;
    if (fseek (f, 0, SEEK_END) == 0) {
        len = ftell(f);
        rewind(f);
    }
    return len;
}

/* create given directory (w/o creating parent directories), return 0 if OK */
/* code from lfs.c */
static int mymakedir(const char* pgmn, lua_State *L, const char* dirname)
{
    int status = 0;
#ifdef WIN32
    int oldmask = umask(0);
    status = _mkdir(dirname);                   /* don't log error here */
#else
    mode_t oldmask = umask( (mode_t)0 );
    mode_t flags = S_IRUSR|S_IWUSR|S_IXUSR|S_IRGRP|S_IWGRP|S_IXGRP|S_IROTH|S_IXOTH;
    status = mkdir(dirname, flags);             /* don't log error here */
#endif
    if (0 == status) status = umask(oldmask);   /* don't log error here */
    return status;
}

/* create given directory and maybe also parent directories, return 0 if OK */
/* code from miniunz.c */
static int mymakedirs(const char* pgmn, lua_State *L, const char* pathname, int isfile)
{
    int status = 0;
    int len = (int)strlen(pathname);
    if (len) {
        char buf[MAX_PATH];
        strcpy(buf, pathname);           /* we change the pathname, so work on copy */
        if (buf[len-1] == '/')
            buf[len-1] = '\0';
        if (mymakedir(pgmn, L, buf)) {  /* try to create in one step, no error if this fails */
            char *p = buf+1;
            while (1) {                 /* failed, so create partwise */
                char hold;
                while(*p && *p != '/')
                    p++;
                hold = *p; *p = 0;
                {
                    int res = mymakedir(pgmn, L, buf);  /* try to create part */
                    if (res && errno == ENOENT) {       /* res <> 0 is not always an error here */
                        status = res;                   /* e.g. if part already exists */
                        break;          /* failed to create one part */
                    }
                }
                if (hold == 0)          /* if last part ... */
                    break;              /* ... leave */
                if (isfile && !strchr(p+1, '/'))
                    break;              /* if pathname is file skip last part */
                *p++ = hold;            /* restore '/' for next sub dir part */
            }
        }   /* else: dir already created in one step */
    }
    return status;
}

/* copy buflen bytes from given file into new file with given name tmpname, return 0 if OK */
static unzFile initzip(const char* pgmn, lua_State *L, char* tmpname, FILE* f, size_t buflen)
{
    FILE* tmpfile = fopen(tmpname, "wb");
    unzFile uf = NULL;
    assert(0 == UNZ_OK);        /* make sure that 0 == OK also for ziplib */
    if (tmpfile) {
#   ifdef USEWIN32IOAPI
        zlib_filefunc_def ffunc;
#   endif
        int status = 0, save_errno = errno;
        errno = 0;
        while(0 == status) {    /* copy zip part of executable in zip file */
            char buf[1024];
            int exp = sizeof(buf);
            int cnt = fread(buf, 1, exp, f);
            if (cnt) {
                exp = cnt;
                cnt = fwrite(buf, 1, exp, tmpfile);
                CHECK0(pgmn, L, status, exp - cnt);
            }
            else
                break;
        }
        errno = save_errno;
        CHECK0(pgmn, L, status, fclose(tmpfile));
        if (0 == status) {      /* no error occured while creating zip file */
#   ifdef USEWIN32IOAPI
            fill_win32_filefunc(&ffunc);
            uf = unzOpen2(tmpname, &ffunc);
#   else
            uf = unzOpen(tmpname);
#   endif
        }
    }
    return uf;
}

/* create given dir and make it the current directory, return 0 if OK */
static int inittempdir(const char* pgmn, lua_State *L, char* dirname)
{
    int status = 0;
    CHECK0(pgmn, L, status, mymakedir(pgmn, L, dirname));
    if (0 == status) CHECK0(pgmn, L, status, chdir(dirname));
    return status;
}

/* extract given zip file to current directory, return 0 if OK */
static int unpackzip(const char* pgmn, lua_State *L, unzFile uf)
{
    int status = 0;
    CHECK0(pgmn, L, status, unzGoToFirstFile(uf));
    while (0 == status) {
        unz_file_info ufi;
        char fn[MAX_PATH+1];
        CHECK0(pgmn, L, status, unzGetCurrentFileInfo(uf, &ufi, fn, sizeof(fn)-1,NULL,0,NULL,0));
        if (status) break;
        if ((ufi.uncompressed_size == 0) && ('/' == fn[strlen(fn)-1])) {    /* directory */
            /* todo: is there a more secure way to identify this as a directory?  */
            /* Note: a directory is not necessarily listed before its files, */
            /* so we don't create dirS here but below on file creation */
        }
        else if ((ufi.uncompressed_size > 0)                                /* file */
                && (CHECK0(pgmn, L, status, unzOpenCurrentFile(uf)) == 0)) {
            int exp = ufi.uncompressed_size;
            char* buf = CHECKNN(pgmn, L, status, malloc(exp));
            if (0 == status) {
                CHECK0(pgmn, L, status, unzReadCurrentFile(uf, buf, exp) - exp);
                if (0 == status) {
                    FILE* nf = fopen(fn, "wb");
                    if ((NULL == nf) && (ENOENT == errno)) {
                        /* can't open file because its parent dir doesn't exist */
                        CHECK0(pgmn, L, status, mymakedirs(pgmn, L, fn, 1));
                        if (0 == status) nf = CHECKNN(pgmn, L, status, fopen(fn, "wb"));
                    }
                    if (0 == status) {
                        CHECK0(pgmn, L, status, fwrite(buf, 1, exp, nf) - exp);
                        CHECK0(pgmn, L, status, fclose(nf));
                    }
                }
            }
            CHECK0(pgmn, L, status, unzCloseCurrentFile(uf));
        }
        if (0 == status) {
            int res_unzGoToNextFile = unzGoToNextFile(uf);
            if (UNZ_END_OF_LIST_OF_FILE == res_unzGoToNextFile) break; /* leave w/o error */
            CHECK0(pgmn, L, status, res_unzGoToNextFile);
        }
    }
    return status;
}

/* copy zip file bytes from given file and extract zip file to temporary directory, return 0 if OK */
static int extractzip(const char* pgmn, lua_State *L, FILE* f, size_t buflen)
{
    int status = 0;
    char* tmpname = CHECKNN(pgmn, L, status, tempnam(NULL, "SRLUA."));
    char* dirname = CHECKNN(pgmn, L, status, tempnam(NULL, "SRLUA"));
    unzFile uf = NULL;
    if (0 == status) {
        status = inittempdir(pgmn, L, dirname);
        if (0 == status) {
            uf = CHECKNN(pgmn, L, status, initzip(pgmn, L, tmpname, f, buflen));
            if (0 == status) {
                status = unpackzip(pgmn, L, uf);
                CHECK0(pgmn, L, status, unzClose(uf));      /* always close zip file */
                CHECK0(pgmn, L, status, remove(tmpname));   /* always remove tmp file */
            }
        }
        if (tmpname) free(tmpname); /* free() has no return value */
        if (dirname) free(dirname); /* tempnam() uses malloc */
    }
    return status;
}

/* run startup code as described for main(), return 0 if OK */
static int handleinit(const char* pgmn, lua_State *L, State *S)
{
    int status = 0;
    const char* init = NULL;
    const char* exec = NULL;
    const char* path = NULL;
    /* always try to read cfg file, since LUA_PATH might be there and not respected otherwise. */
    /* Note: files are already extracted and directory is set here, so just open */
    FILE* cfg = fopen(LUA_INIT_CFGFILE, "r");   /* no error, if not existing */
    if (cfg) {
        fclose(cfg);            /* file exists, just close */
        status = DOFILE(pgmn, L, LUA_INIT_CFGFILE);
        if (0 == status) {      /* file parsed w/o error, get config settings */
            lua_getglobal(L, "init");
            init = lua_tostring(L, -1); lua_pop(L, 1);  /* might still be NULL */
            lua_getglobal(L, "exec");
            exec = lua_tostring(L, -1); lua_pop(L, 1);  /* might still be NULL */
            /* no special processing for LUA_PATH required, if it is set in the cfg file */
            /* it will just be set as a global variable. */
            /* eval sequence: global var (<= cfg file), environment var, LUA_PATH_DEFAULT */
        }
    }
    if (0 == status) {          /* let environment variables override config file settings */
#define OVR(cvar, envvar) { const char* tmp = getenv(#envvar); if (tmp) cvar = tmp; }
        OVR(init, LUA_INIT);
        OVR(exec, LUA_EXEC);
        OVR(path, LUA_PATH);
#undef OVR
    }
#ifdef LUA_INIT_DEFAULT
    if (0 == status) {
        if ((init == NULL) && strlen(LUA_INIT_DEFAULT)) {
            FILE* tf;
            if (LUA_INIT_DEFAULT[0] == '@') {
                if ((tf = fopen(LUA_INIT_DEFAULT+1,"rb")) != NULL) {
                    fclose(tf);     /* we just checked for existence */
                    init = LUA_INIT_DEFAULT;
                }   /* else: default init file does not exist, so don't use it */
            }
            else    /* else: init string doesn't defined a file, but a chunk */
                init = LUA_INIT_DEFAULT;
        }
    }
#endif
#ifdef LUA_EXEC_DEFAULT
    if (0 == status) {
        if ((exec == NULL) && strlen(LUA_EXEC_DEFAULT)) {
            FILE* tf;
            if ((tf = fopen(LUA_EXEC_DEFAULT,"rb")) != NULL) {
                fclose(tf);         /* we just checked for existence */
                exec = LUA_EXEC_DEFAULT;
            }   /* else: default exec file does not exist, so don't use it */
        }
    }
#endif
    if (0 == status) {
        if (init) {
            if (init[0] == '@')
                status = DOFILE(pgmn, L, (char*)init+1);
            else
                status = DOSTRING(pgmn, L, init, "=LUA_INIT");
        }
        if (exec) {
            FILE *f = CHECKNN(pgmn, L, status, fopen(exec,"rb"));
            if (0 == status) {
                size_t size = myfilelen(f);
                if (size) {
                    assert(&S);
                    S->f = f; S->size = size;
                }
            }
        }
    }
    return status;
}

static int load(lua_State *L, const char* pgmn)
{
    Glue t;
    State S = {NULL, 0, '\0'};
    int status = 0;
    FILE *f = CHECKNN(pgmn, L, status, fopen(pgmn,"rb"));
    if (0 == status) {
        CHECK0(pgmn, L, status, fseek(f,-sizeof(t),SEEK_END));
        if (0 == status) CHECK0(pgmn, L, status, fread(&t, sizeof(t), 1, f) - 1);
        if (0 == status) CHECK0(pgmn, L, status, ("Not glued.", memcmp(t.sig,GLUESIG,GLUETYP)));
        if (0 == status) CHECK0(pgmn, L, status, fseek(f,t.size1,SEEK_SET));
    }
    if (0 == status) {
        switch(t.sig[GLUETYP]) {
            case 'L': {
                S.f=f; S.size=t.size2;
                CHECK0(pgmn, L, status, lua_load(L,myget,&S,pgmn));
                break;
            }
            case 'Z': {
                if (0 == status) CHECK0(pgmn, L, status, fseek(f,t.size1,SEEK_SET));
                if (0 == status) extractzip(pgmn, L, f, t.size2);
                if (0 == status) handleinit(pgmn, L, &S);
                /* S now contains info for file to be initially executed */
                if (0 == status && S.f && S.size) {
                    /* lua_load() will push a function on the stack that is called */
                    /* by lua_call() later on in pmain(), so ... */
                    CHECK0(pgmn, L, status, lua_load(L,myget,&S,pgmn));
                    /* if an error occurs in the script defined by S, control is */
                    /* immediately passed back to lua_cpcall()! */
                }
                else    /* ... return error status if we didn't call lua_load() */
                    CHECK0(pgmn, L, status, ("Can't run startup code.", -1));
                break;
            }
            default:    /* unknown glue tag, so return error status */
                CHECK0(pgmn, L, status, ("Unknown glue sig.", -1));
        }
    }
    return status;
}

static int pmain(lua_State *L)
{
    int status = 0;
    char **argv=lua_touserdata(L,1);
    openlibs(L);
    status = load(L,argv[0]);
    if (0 == status) {
        int i;
        lua_pushliteral(L,"arg");
        lua_newtable(L);
        for (i=0; argv[i]; i++) {
            lua_pushstring(L,argv[i]);
            lua_rawseti(L,-2,i);
        }
        lua_pushliteral(L,"n");
        lua_pushnumber(L,i-1);
        lua_rawset(L,-3);
        lua_rawset(L,LUA_GLOBALSINDEX);
        lua_call(L,0,0);
    }
    return status;
}

/**
 * Short instructions:
 *
 * To create a self running executable:
 * - one single lua file:
 *   glue srlua.exe foo.lua foo.exe
 * - multiple files stored in one zip file:
 *   glue srlua.exe bar.zip bar.exe Z
 *
 * 'glue' is the unchanged executable as provided by the srlua package.
 * 'srlua' is the executable created from this C module.
 *
 * How to work w/ the zip file version:
 *   Just as you would do w/ the original srlua executable ;o).
 *
 * How to work w/ the zip file version:
 * - Put a file 'config.lua' (see macro definition for 'LUA_INIT_CFGFILE')
 *   into the root dir of the zip file.
 *   This file might define the following variables:
 *   - 'init': A string that defines either the name of an initial lua file
 *             or a lua chunk containing startup code.
 *             If it defines a lua file name, the string must start w/ '@'.
 *             If environment variable 'LUA_INIT' is set, it will override
 *             the definition of 'init'.
 *             If environment variable 'LUA_INIT' is not set and 'init' is
 *             defined the macro definition 'LUA_INIT_DEFAULT' is used
 *             (which might be empty ("" or undefined).
 *             In this case if 'LUA_INIT_DEFAULT' defines a file name this
 *             file name will only be used if the file exists in the
 *             root dir of the zip file.
 *             If no init file is defined, no startup code is executed.
 *   - 'exec': A string that defines the name of the lua file containing the
 *             'main' code to be started (for sure other files contained in
 *             zip can be used, e.g. via 'require').
 *             Since 'exec' always defines a lua file name, it does not start
 *             w/ '@'.
 *             If environment variable 'LUA_EXEC' is set, it will override
 *             the definition of 'exec'.
 *             If environment variable 'LUA_EXEC' is not set and 'exec' is
 *             defined the macro definition 'LUA_EXEC_DEFAULT' is used
 *             (which might be empty ("" or undefined).
 *             In this case the file name defined by 'LUA_EXEC_DEFAULT' will
 *             only be used if the file exists in the root dir of the zip file.
 *             Since an executable lua source is required in order to perform
 *             anything w/ this executable, the exec file must always be
 *             defined or an error will occur.
 *   - 'LUA_PATH': A string using the standard 'LUA_PATH' semantics.
 *             If environment variable 'LUA_PATH' is set, it will override
 *             the definition of 'LUA_PATH' from within the config file.
 *             Note that since all lua code is extracted from the zip file
 *             contained in this executable the 'LUA_PATH' should contain
 *             relative path names to the sources in the zip file.
 *
 * Notes:
 * - Whereas the temporary zip file extracted from this executable will be
 *   deleted before the program quits, the temporary directory extracted
 *   from the zip file is currently *not* automatically removed.
 * - No command line argument processing as provided by the standard lua
 *   standalone executable is done here. Arguments are just pushed on the
 *   stack before the 'exec' file is started, processing these arguments
 *   is than a task of the application logic.
 *
 * TODO:
 * - remove temporary unzip directory
 * - support script from stdin and interactive mode after start
 *   (flags '-' and '-i' from standard lua standalone)
 * - see todoS in code
 * - add IUP / wxLua startup code
 **/
int main(int argc, char *argv[])
{
    lua_State *L;
#ifdef _WIN32
    char name[MAX_PATH];
    if (GetModuleFileName(NULL,name,sizeof(name))==0) {
        MESSAGE(argv[0], NULL, NULL, "Can't locate this executable.");
        return EXIT_FAILURE;
    }
    argv[0]=name;
#endif
    L=lua_open();
    if (L==NULL) {
        MESSAGE(argv[0], NULL, NULL, "Can't allocate for state.");
        return EXIT_FAILURE;
    }
    if (lua_cpcall(L,pmain,argv)) FATAL(argv[0], L, NULL, lua_tostring(L,-1));
    lua_close(L);
    return EXIT_SUCCESS;
}
